from valiml.boosting.DecisionStump import DecisionStump
from valiml.boosting.MulticlassDecisionStump import DecisionStumpSamme, DecisionStumpSammeR

from valiml.boosting.AdaBoost import AdaBoost

__all__ = [
    'AdaBoost',
    'DecisionStumpSamme',
    'DecisionStumpSammeR'
]



import numba
import numpy as np
from numba import prange
from sklearn.base import BaseEstimator, ClassifierMixin


@numba.jit(nopython=True, cache=True, parallel=True)
def fit_stump(X, Y, feature_ids, sample_weights=None):
    n_features = len(feature_ids)

    weighted_errors = np.zeros((n_features,), dtype=np.float64)
    split_values = np.zeros((n_features,), dtype=np.float64)
    labels = np.zeros((n_features,), dtype=np.int32)

    for fidx in prange(n_features):
        feature_idx = feature_ids[fidx]

        sorted_data_mask = X[:, feature_idx].argsort()
        x = X[:, feature_idx][sorted_data_mask]
        y = Y[sorted_data_mask]
        sample_distribution = np.cumsum((2 * y - 1) * sample_weights[sorted_data_mask])

        best_weighted_error = 0.5 - np.abs(sample_distribution[-1]) / 2
        split_value = x[0]
        label = 1 if sample_distribution[-1] < 0 else -1

        for idx in range(1, y.shape[0]):
            if x[idx] != x[idx - 1]:
                margin = sample_distribution[-1] - 2 * sample_distribution[idx - 1]
                weighted_error = 0.5 - np.abs(margin) / 2
                if weighted_error < best_weighted_error:
                    split_value = x[idx - 1] + (x[idx] - x[idx - 1]) / 2
                    label = 1 if margin < 0 else -1
                    best_weighted_error = weighted_error

        weighted_errors[fidx] = best_weighted_error
        split_values[fidx] = split_value
        labels[fidx] = label

    best_idx = weighted_errors.argmin()
    return weighted_errors[best_idx], feature_ids[best_idx], split_values[best_idx], labels[best_idx]


class DecisionStump(BaseEstimator, ClassifierMixin):
    def __init__(self, n_random_features=None):
        self.error = 1
        self.feature_idx = -1
        self.split_value = 0
        self.label = 0
        self.n_random_features = n_random_features

    def fit(self, X, y, sample_weight=None):
        if self.n_random_features is not None:
            n_random_features = min(self.n_random_features, X.shape[1])
            features_subset = np.random.choice(X.shape[1], n_random_features, replace=False)
        else:
            features_subset = list(range(X.shape[1]))

        error, self.feature_idx, self.split_value, self.label = fit_stump(X, y, features_subset,
                                                                          sample_weights=sample_weight)
        return self

    def decision_function(self, X):
        return np.sign(2 * self.predict(X) - 1)

    def predict(self, X):
        if self.label == 1:
            return (X[:, self.feature_idx] < self.split_value).astype('int')
        return (X[:, self.feature_idx] >= self.split_value).astype('int')

    def __lt__(self, other):
        return self.split_value < other.split_value



import numba
import numpy as np
import pandas as pd
from numba import prange
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.multiclass import OneVsRestClassifier
from sklearn.preprocessing import MinMaxScaler

from valiml.utils.numba_utils import create_progbar_numba


@numba.jit(nopython=True, parallel=True)
def optimize(X, y, alpha, max_iters):
    t = 0
    w = np.zeros((X.shape[1] + 1,), dtype=np.float32)
    bar = create_progbar_numba(max_iters)
    for n_iter in range(max_iters):
        for idx in range(X.shape[0]):
            t += 1
            lr = 1 / (t * alpha)
            if y[idx] * np.dot(w[1:], X[idx, :]) < 1:
                w[1:] -= lr * (alpha * w[1:] - y[idx] * X[idx])
            else:
                w[1:] -= lr * alpha * w[1:]
        bar.add(1)
    return w


def make_compute_kernel_matrix(kernel):
    @numba.jit(nopython=True, parallel=True)
    def _compute(X, kernel_matrix):
        for idx in range(X.shape[0]):
            kernel_matrix[idx:, idx] = kernel(X[idx:, :], X[idx, :])
            kernel_matrix[idx, idx:] = kernel_matrix[idx:, idx]

        return kernel_matrix
    return _compute


@numba.jit(nopython=True, parallel=True)
def optimize_kernel(X, y, alpha, max_iters, kernel_matrix):
    t = 0
    a = np.zeros((X.shape[0]), dtype=np.float32)
    bar = create_progbar_numba(max_iters)
    for n_iter in range(max_iters):
        for idx in range(X.shape[0]):
            t += 1
            lr = 1 / (t * alpha)

            value = lr * y[idx] * (a * y * kernel_matrix[idx, :]).sum()
            if value < 1:
                a[idx] = a[idx] + 1
        bar.add(1)
    return a


def make_rbf(gamma):
    @numba.jit(nopython=True, parallel=True)
    def rbf(X, w):
        result = np.zeros((X.shape[0],), dtype=np.float32)
        for idx in prange(X.shape[0]):
            result[idx] = -np.linalg.norm(X[idx, :] - w)**2 / (2 * gamma**2)
        return np.exp(result)
    return rbf


def make_apply_kernel(kernel):
    @numba.jit(nopython=True, parallel=True)
    def _kernel(X_train, X_test, a):
        result = np.zeros((X_test.shape[0],), dtype=np.float32)
        for idx in range(X_test.shape[0]):
            result[idx] = (kernel(X_train, X_test[idx]) * a).sum()
        return result
    return _kernel


class HingeSVM(BaseEstimator, ClassifierMixin):
    def __init__(self, optimizer='pegasos', max_iters=1000, alpha=0.1, kernel='linear', verbose=0, gamma=0.1):
        """

        :param optimizer: The optimizer to use. Choices include lm and gd.
        :param max_iters: Maximum number of iterations for convergence.
        :param alpha: The weight of the squared norm of w.
        :param n_inits: The number of initial starting position to try.
        :param regularizer: The regularizer of the respective optimizer.
                            Step size for gd and l2 regularization for lm.
        """
        self.optimizer = optimizer
        self.max_iters = max_iters
        self.alpha = alpha
        self.kernel = kernel
        self.verbose = verbose
        self.gamma = gamma

    def fit(self, X, y):
        self.coef_ = np.zeros((1, X.shape[1]))
        self.intercept_ = np.zeros((1, 1))

        X = X.astype('float32')
        y = 2 * y - 1
        if self.kernel == 'linear':
            w = optimize(X, y, self.alpha, self.max_iters)

            self.intercept_[0] = w[0]
            self.coef_[:] = w[1:]

            hinge_loss = np.clip(1 - self.decision_function(X) * y, 0, np.inf)
            self.support_vectors_ = X[hinge_loss > 0]
            self.dual_coef_ = hinge_loss[hinge_loss > 0]

        elif self.kernel == 'rbf':
            kernel = make_rbf(self.gamma)
            kernel_matrix = np.zeros((X.shape[0], X.shape[0]), dtype='float32')
            make_compute_kernel_matrix(kernel)(X, kernel_matrix)
            self.a = optimize_kernel(X, y, self.alpha, self.max_iters, kernel_matrix) * y
            self.X_train = X

        return self

    def decision_function(self, X):
        X = X.astype('float32')
        if self.kernel == 'linear':
            result = np.matmul(X, self.coef_[0])
        elif self.kernel == 'rbf':
            kernel = make_rbf(self.gamma)
            result = make_apply_kernel(kernel)(self.X_train, X, self.a)
        return result

    def predict(self, X):
        return np.sign(self.decision_function(X))


@numba.jit(nopython=True)
def test_progb():
    bar = create_progbar_numba(25)
    a = np.random.rand(2000, 10000)
    b = np.random.rand(10000, 2000)
    c = 0
    for idx in range(25):
        c += np.sum(a + b.T)

        bar.add(1)
    return c



import numba
import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.ensemble import AdaBoostClassifier
from sklearn.multiclass import OneVsRestClassifier
from sklearn.utils import check_array, check_X_y

from valiml.utils.numba_utils import create_progbar_numba

from valiml.boosting.DecisionStump import DecisionStump
from valiml.boosting.MulticlassDecisionStump import DecisionStumpSamme, DecisionStumpSammeR
from valiml.optimizers import lm
from sklearn.model_selection import train_test_split


# TODO: Think of a way to use inheritance for multiple modes
from valiml.utils.utils import normalize


class AdaBoost(BaseEstimator, ClassifierMixin):
    def __init__(self, n_estimators=200, n_random_features=None, mode='exponential', type='discrete', verbose=False):
        self.n_estimators = n_estimators
        self.mode = mode
        self.n_random_features = n_random_features
        self.type = type
        self.verbose = verbose
        
        if self.mode == 'exponential':
            self._Z = 1
            
    def _get_weak_classifier(self, instance_importance, x, y):
        if self.n_labels > 2 and self.mode == 'exponential':
            if self.type == 'discrete':
                return DecisionStumpSamme(n_random_features=self.n_random_features)\
                    .fit(x, y, sample_weights=instance_importance)
            else:
                return DecisionStumpSammeR(n_random_features=self.n_random_features)\
                    .fit(x, y, sample_weight=instance_importance)
        elif self.n_labels > 2 and self.mode == 'quadratic':
            return DecisionStump(n_random_features=self.n_random_features)\
                .fit(x, y, sample_weight=instance_importance)
        else:
            return DecisionStump(n_random_features=self.n_random_features)\
                .fit(x, y, sample_weight=instance_importance)
            
    def _get_updated_instance_importance(self, instance_importance, weighted_error,
                                         alpha, stump_prediction, classifier_decision, y):
        missclassified_samples = (stump_prediction != y)
        if self.mode == 'exponential':
            if self.n_labels > 2:
                if self.type == 'discrete':
                    return normalize(instance_importance * np.exp(missclassified_samples * alpha))
                else:
                    result = np.array([np.dot(self._y[idx], self._log_proba[idx]) for idx in range(self._y.shape[0])])
                    return normalize(instance_importance * np.exp(-result * (self.n_labels - 1) / self.n_labels))
            else:
                instance_importance[missclassified_samples] /= 2 * weighted_error
                instance_importance[~missclassified_samples] /= 2 * (1 - weighted_error)
                return instance_importance
        if self.mode == 'quadratic':
            if self.n_labels > 2:
                return 2 * (classifier_decision - y) / y.shape[0]
            else:
                return - 2 * (1 - classifier_decision * y)
        elif self.mode == 'logit':
            return normalize(- 1 / (1 + np.exp(classifier_decision * y)) / y.shape[0])

    def _get_alpha(self, n_iter, instance_importance, weighted_error,
                   stump_decision, previous_classifier_decision, y):
        if self.mode == 'exponential':
            if self.n_labels > 2:
                if self.type == 'discrete':
                    return np.log((1 - weighted_error) / weighted_error) + np.log(self.n_labels - 1)
                else:
                    return 1
            else:
                return np.log((1 - weighted_error) / weighted_error) / 2
        elif self.mode == 'quadratic':
            if self.n_labels > 2:
                return ((y - previous_classifier_decision) * stump_decision).sum() / (stump_decision != 0).sum()
            else:
                return ((y - previous_classifier_decision) * stump_decision).sum() / y.shape[0]
        elif self.mode == 'logit':
            previous_sum = previous_classifier_decision * y
            stump = stump_decision * y

            def logit(alpha, jacobian=False):
                loss = np.log(1 + np.exp(-previous_sum - alpha * stump)) / y.shape[0]
                if jacobian:
                    derivative = (-(1 / (1 + np.exp(previous_sum + alpha * stump))) * stump / y.shape[0]).sum()
                    return np.array([loss.sum()]), np.array([derivative])
                return np.array([loss.sum()])

            return lm(np.array([0]), logit, max_iters=100, alpha=1e1)[0][0]
        
    def _compute_loss(self, weighted_error, classifier_decision, y):
        if self.mode == 'exponential':
            if self.n_labels > 2:
                return np.exp((-1 / self.n_labels) * (self._y * classifier_decision).sum(axis=1)).sum() / y.shape[0]
            if self.n_labels == 2:
                self._Z *= 2 * np.sqrt(weighted_error * (1 - weighted_error))
                return self._Z
        elif self.mode == 'quadratic':
            if self.n_labels > 2:
                return ((classifier_decision - y) ** 2).sum() / y.shape[0]
            else:
                return ((1 - y * classifier_decision) ** 2).sum() / y.shape[0]
        elif self.mode == 'logit':
            return (np.log(1 + np.exp(-classifier_decision * y))).sum() / y.shape[0]
    
    def fit(self, X, Y, validation_split=None):
        """
        :param X: features
        :param Y: labels
        """
        X, Y = check_X_y(X, Y, multi_output=False, y_numeric=True, estimator='AdaBoost')

        check_array(X, accept_sparse=False, dtype='numeric', force_all_finite=True, estimator='AdaBoost')
        check_array(Y, accept_sparse=False, dtype='numeric', force_all_finite=True, ensure_2d=False,
                    estimator='AdaBoost')

        if self.n_random_features is None:
            self.n_random_features = X.shape[1]

        if self.verbose:
            metrics = ['w_err', 'loss', 'err', 'acc']
            if validation_split is not None:
                metrics.append('val_acc')
            self._bar = create_progbar_numba(self.n_estimators, stateful_metrics="|".join(metrics))

        if validation_split is not None:
            if isinstance(validation_split, tuple):
                x_test, y_test = validation_split
            else:
                X, x_test, Y, y_test = train_test_split(X, Y, test_size=validation_split, shuffle=True)
        
        self.losses = []
        self.errors = []
        self.weighted_errors = []
        self.decision_stumps = []
        self.instance_importance = np.ones((X.shape[0],)) / X.shape[0]
        if validation_split is not None:
            self.validation_errors = []

        self.n_labels = len(np.unique(Y))
        if self.n_labels > 2 and self.mode == 'exponential':
            classifier_decision = np.zeros((X.shape[0], self.n_labels))
            if validation_split is not None:
                validation_decision = np.zeros((x_test.shape[0], self.n_labels))
        else:
            classifier_decision = np.zeros((X.shape[0],))
            if validation_split is not None:
                validation_split = np.zeros((x_test.shape[0],))

        if self.n_labels > 2 and self.mode == 'exponential':
            self._y = np.zeros((Y.shape[0], self.n_labels))
            self._y.fill(-1/(self.n_labels-1))
            for idx in range(Y.shape[0]):
                self._y[idx, Y[idx]] = 1
        elif self.n_labels > 2 and self.mode == 'quadratic':
            Y = 2 * Y

        n_iter = 1
        while n_iter <= self.n_estimators:
            decision_stump = self._get_weak_classifier(self.instance_importance, X, Y)
            stump_prediction = decision_stump.predict(X)
            weighted_error = self.instance_importance[stump_prediction != Y].sum()

            if weighted_error == 0:
                self.decision_stumps.append((1.0, decision_stump))
                if self.verbose:
                    values = [0, 0, 0, 1]

                    self._bar.update(
                        self.n_estimators,
                        "|".join(metrics),
                        values
                    )

                break
            elif weighted_error == 0.5 and self.n_labels <= 2:
                break

            stump_decision = decision_stump.decision_function(X)
            if self.n_labels > 2 and self.mode == 'exponential' and self.type != 'discrete':
                self._log_proba = np.log(decision_stump.predict_proba(X))

            self.weighted_errors.append(weighted_error)

            alpha = self._get_alpha(n_iter,
                                    self.instance_importance,
                                    weighted_error,
                                    stump_decision,
                                    classifier_decision,
                                    Y)
            if not (self.n_labels > 0 and self.mode == 'quadratic'):
                assert alpha >= 0
            self.decision_stumps.append((alpha, decision_stump))

            classifier_decision += alpha * stump_decision
            if validation_split is not None:
                validation_decision += alpha * decision_stump.decision_function(x_test)

            classifier_error = (self._predict(classifier_decision) != Y).sum() / Y.shape[0]
            if validation_split is not None:
                validation_error = (self._predict(validation_decision) != y_test).sum() / y_test.shape[0]

            self.errors.append(classifier_error)
            if validation_split is not None:
                self.validation_errors.append(validation_error)

            self.instance_importance = self._get_updated_instance_importance(
                self.instance_importance,
                weighted_error,
                alpha,
                stump_prediction,
                classifier_decision,
                Y
            )

            loss = self._compute_loss(weighted_error, classifier_decision, Y)
            self.losses.append(loss)
            if self.verbose:
                values = [
                    weighted_error,
                    loss,
                    classifier_error,
                    1 - classifier_error
                ]
                if validation_split is not None:
                    values.append(1 - validation_error)

                self._bar.update(
                    n_iter,
                    "|".join(metrics),
                    values
                )

            n_iter += 1

        return self

    def decision_function(self, X):
        if self.n_labels > 2 and self.mode == 'exponential':
            prediction = np.zeros((X.shape[0], self.n_labels))
        else:
            prediction = np.zeros((X.shape[0],))

        for alpha, decision_stump in self.decision_stumps:
            prediction += alpha * decision_stump.decision_function(X)

        return prediction

    def _predict(self, decision):
        if self.n_labels > 2 and self.mode == 'exponential':
            return decision.argmax(axis=1)
        elif self.n_labels == 2:
            return (decision > 0).astype('int')
        else:
            return NotImplemented()

    def predict(self, X):
        decision = self.decision_function(X)
        return self._predict(decision)

    def predict_proba(self, X):
        if self.mode == 'exponential':
            if self.n_labels == 2:
                alpha_sum = [alpha for alpha, prediction in self.decision_stumps]
                return (self.decision_function(X) + alpha_sum) / (2 * alpha_sum)
            else:
                if self.type == 'discrete':
                    raise NotImplemented()
                else:
                    prediction = self.decision_function(X)
                    prediction /= self.n_labels
                    prediction = np.exp((1. / (self.n_labels - 1)) * prediction)

                    normalizer = prediction.sum(axis=1)[:, np.newaxis]
                    normalizer[normalizer == 0.0] = 1.0
                    prediction /= normalizer
                    return prediction



from os.path import dirname, join
from cffi import FFI
ffibuilder = FFI()

__dirname__ = dirname(__file__)

ffibuilder.set_source(
    "_utils",
    open(join(__dirname__, "utils.c")).read(),
    sources=[],
    libraries=["c"],
)

with open(join(__dirname__, "utils.h")) as f:
    ffibuilder.cdef(f.read())

if __name__ == "__main__":
    ffibuilder.compile()



from copy import deepcopy
import numpy as np
from scipy.stats import mode
from itertools import combinations

from sklearn.base import BaseEstimator, ClassifierMixin, clone


def one_vs_rest_split(X, Y):
    classes = np.unique(Y)
    for class_val in classes:
        yield class_val, X, (Y == class_val).astype('int')


def one_vs_one_split(X, Y):
    classes = np.unique(Y)
    for class_vals in combinations(classes, 2):
        mask = np.isin(Y, class_vals)
        yield class_vals, X[mask], 2 * (Y[mask] == class_vals[0]) - 1


class OneVsRestClassifier(BaseEstimator, ClassifierMixin):
    def __init__(self, classifier, verbose=0):
        self.classifier = classifier
        self.verbose = verbose
    
    def fit(self, X, Y, *args, **kwargs):
        self.classifiers = {}
        self.classes = np.unique(Y)
        for class_val, x, y in one_vs_rest_split(X, Y):
            classifier = clone(self.classifier)
            classifier.fit(x, y, *args, **kwargs)
            self.classifiers[class_val] = classifier
            if self.verbose:
                print(f'Trained classifier for class {class_val}')
        return self
    
    def predict(self, x):
        predictions = np.zeros((x.shape[0], len(self.classifiers)))
        for idx, class_vals in enumerate(self.classifiers):
            classifier = self.classifiers[class_vals]
            predictions[:, idx] = classifier.decision_function(x)
        
        return np.vectorize(lambda x: self.classes[x])(predictions.argmax(axis=1)).reshape(-1)

    
class OneVsOneClassifier():
    def __init__(self, classifier):
        self.classifier = classifier
    
    def fit(self, X, Y, *args, **kwargs):
        self.classifiers = {}
        self.classes = np.unique(Y)
        for class_vals, x, y in one_vs_one_split(X, Y):
            classifier = deepcopy(self.classifier)
            classifier.fit(x, y, *args, **kwargs)
            self.classifiers[class_vals] = classifier
        return self
    
    def predict(self, x):
        predictions = np.zeros((x.shape[0], len(self.classifiers)))
        for idx, class_vals in enumerate(self.classifiers):
            classifier = self.classifiers[class_vals]
            prediction = classifier.predict(x)
            predictions[:, idx] = np.vectorize(lambda value: class_vals[0 if value > 0 else 1])(prediction)
        
        return mode(predictions, axis=1).mode.reshape(-1)


import numpy as np


# LEVENBERG - MARQUARDT
def lm(start_x, function, max_iters=10, alpha=1e9, max_iters_no_change=30, callback=None):
    n_input_dimensions = start_x.shape[0]
    y, jacobian = function(start_x, jacobian=True)
    n_output_dimensions = y.shape[0]
    
    A = np.zeros((n_output_dimensions + n_input_dimensions, n_input_dimensions))
    solution = np.zeros((n_output_dimensions + n_input_dimensions,))
    
    x = start_x
    n_iter = 1
    last_change_iter = 1
    while n_iter <= max_iters:
        A[:n_output_dimensions, :] = jacobian
        A[n_output_dimensions:, :] = np.sqrt(alpha) * np.identity(n_input_dimensions)
        solution[:n_output_dimensions] = -y
    
        lstsq_solution = np.linalg.lstsq(A, solution, rcond=None)[0]
        next_x = x + lstsq_solution
        next_y, new_jacobian = function(next_x, jacobian=True)
        
        if n_iter - last_change_iter > max_iters_no_change:
            break

        if next_y < y:
            x = next_x
            y = next_y
            alpha *= 0.8
            jacobian = new_jacobian
            last_change_iter = n_iter
        else:
            alpha *= 2.0

        if callback is not None:
            callback(n_iter, x, y, alpha)
            
        n_iter += 1
    
    return x, y


def gd_online(X, start_x, cost_function, max_iters=1000, step_size=0.1, max_iters_no_change=40, callback=None):
    y, jacobian = cost_function(start_x, 0, jacobian=True)

    is_step_size_function = callable(step_size)
    n_pass_iter = 1
    n_iter = 1
    x = start_x
    iter_step_size = step_size
    while n_pass_iter <= max_iters:
        for idx in range(X.shape[0]):
            iter_step_size = step_size(n_iter) if is_step_size_function else iter_step_size
            if len(jacobian.shape) == 1:
                next_x = x - iter_step_size * jacobian
            else:
                next_x = x - iter_step_size * jacobian.mean(axis=1)

            next_y, new_jacobian = cost_function(next_x, idx, jacobian=True)

            x = next_x
            y = next_y
            jacobian = new_jacobian

            if callback is not None:
                callback(n_iter, x, y, iter_step_size)

            n_iter += 1

        n_pass_iter += 1
    return x, y


def gd(start_x, cost_function, max_iters=1000, step_size=0.1, max_iters_no_change=40, callback=None):
    y, jacobian = cost_function(start_x, jacobian=True)

    is_step_size_function = callable(step_size)
    last_change_iter = 1
    n_iter = 1
    x = start_x
    iter_step_size = step_size
    while n_iter <= max_iters:
        iter_step_size = step_size(n_iter) if is_step_size_function else iter_step_size
        if len(jacobian.shape) == 1:
            next_x = x - iter_step_size * jacobian
        else:
            next_x = x - iter_step_size * jacobian.mean(axis=1)

        next_y, new_jacobian = cost_function(next_x, jacobian=True)

        if n_iter - last_change_iter > max_iters_no_change:
            break

        if next_y < y:
            x = next_x
            y = next_y
            if not is_step_size_function:
                step_size *= 2
            jacobian = new_jacobian
            last_change_iter = n_iter
        else:
            if not is_step_size_function:
                iter_step_size *= 0.8

        if callback is not None:
            callback(n_iter, x, y, step_size)

        n_iter += 1

    return x, y



import numba
import numpy as np
from numba import prange
from sklearn.base import BaseEstimator, ClassifierMixin


@numba.jit(nopython=True, cache=True)
def normalize(v):
    return v / v.sum()


@numba.jit(nopython=True, parallel=False, cache=True)
def fit_stump_samme(X, Y, feature_ids, n_labels, sample_weight=None):
    n_features = len(feature_ids)

    weighted_errors = np.zeros((n_features,), dtype=np.float64)
    split_values = np.zeros((n_features,), dtype=np.float64)
    left_labels = np.zeros((n_features,), dtype=np.int32)
    right_labels = np.zeros((n_features,), dtype=np.int32)

    for fidx in prange(n_features):
        feature_idx = feature_ids[fidx]

        sorted_data_mask = X[:, feature_idx].argsort()
        x = X[:, feature_idx][sorted_data_mask]
        y = Y[sorted_data_mask]
        sorted_sample_weights = sample_weight[sorted_data_mask]

        sample_sum_left = np.zeros(n_labels)
        sample_sum_right = np.zeros(n_labels)
        for idx in range(y.shape[0]):
            sample_sum_right[y[idx]] += sorted_sample_weights[idx]

        total_weighted_sum = sample_sum_right.sum()

        best_split_value = x[0]
        best_left_label = np.random.choice(n_labels)
        best_right_label = sample_sum_right.argmax()
        best_weighted_accuracy = sample_sum_right[best_right_label]

        for idx in range(y.shape[0]):
            if idx > 0 and x[idx - 1] != x[idx]:
                left_label = sample_sum_left.argmax()
                right_label = sample_sum_right.argmax()
                weighted_accuracy = sample_sum_left[left_label] + sample_sum_right[right_label]

                if weighted_accuracy > best_weighted_accuracy:
                    best_split_value = x[idx - 1] + (x[idx] - x[idx - 1]) / 2
                    best_left_label = left_label
                    best_right_label = right_label
                    best_weighted_accuracy = weighted_accuracy

            sample_sum_left[y[idx]] += sorted_sample_weights[idx]
            sample_sum_right[y[idx]] -= sorted_sample_weights[idx]

        weighted_errors[fidx] = total_weighted_sum - best_weighted_accuracy
        split_values[fidx] = best_split_value
        left_labels[fidx] = best_left_label
        right_labels[fidx] = best_right_label

    best_idx = weighted_errors.argmin()

    return (weighted_errors[best_idx], feature_ids[best_idx],
            split_values[best_idx],
            left_labels[best_idx], right_labels[best_idx])


@numba.jit(nopython=True, parallel=True, cache=True)
def fit_stump_sammer(X, Y, feature_ids, n_labels, sample_weight=None):
    n_features = len(feature_ids)

    weighted_errors = np.zeros((n_features,), dtype=np.float64)
    split_values = np.zeros((n_features,), dtype=np.float64)
    left_probabilities = np.zeros((n_features, n_labels), dtype=np.float64)
    right_probabilities = np.zeros((n_features, n_labels), dtype=np.float64)

    for fidx in prange(n_features):
        feature_idx = feature_ids[fidx]

        sorted_data_mask = X[:, feature_idx].argsort()
        x = X[:, feature_idx][sorted_data_mask]
        y = Y[sorted_data_mask]
        sorted_sample_weights = sample_weight[sorted_data_mask]

        sample_sum_left = np.zeros(n_labels)
        sample_sum_right = np.zeros(n_labels)
        for idx in range(0, y.shape[0]):
            sample_sum_right[y[idx]] += sorted_sample_weights[idx]
        total_weighted_sum = sample_sum_right.sum()

        best_split_value = x[0]
        best_left_probabilities = normalize(np.random.random((n_labels,)))
        best_right_probabilities = normalize(sample_sum_right)
        best_weighted_accuracy = sample_sum_right[sample_sum_right.argmax()]

        for idx in range(0, y.shape[0]):
            if idx > 0 and x[idx - 1] != x[idx]:
                if x[idx] != best_split_value:
                    left_label = sample_sum_left.argmax()
                    right_label = sample_sum_right.argmax()
                    weighted_accuracy = sample_sum_left[left_label] + sample_sum_right[right_label]

                    if weighted_accuracy > best_weighted_accuracy:
                        best_split_value = x[idx - 1] + (x[idx] - x[idx - 1]) / 2
                        best_left_probabilities = normalize(sample_sum_left)
                        if sample_sum_right.sum() == 0:
                            best_right_probabilities = normalize(np.random.random((n_labels,)))
                        else:
                            best_right_probabilities = normalize(sample_sum_right)
                        best_weighted_accuracy = weighted_accuracy

            sample_sum_left[y[idx]] += sorted_sample_weights[idx]
            sample_sum_right[y[idx]] -= sorted_sample_weights[idx]

        weighted_errors[fidx] = total_weighted_sum - best_weighted_accuracy
        split_values[fidx] = best_split_value
        left_probabilities[fidx] = best_left_probabilities
        right_probabilities[fidx] = best_right_probabilities

    best_idx = weighted_errors.argmin()
    return (weighted_errors[best_idx], feature_ids[best_idx],
            split_values[best_idx],
            left_probabilities[best_idx], right_probabilities[best_idx])


class DecisionStumpSamme(BaseEstimator, ClassifierMixin):
    def __init__(self, n_random_features=None):
        self.feature_idx = -1
        self.left_label = 0
        self.right_label = 0
        self.split_value = 0
        self.n_labels = 0
        self.classes_ = 0
        self.n_random_features = n_random_features

    def fit(self, X, y, sample_weights=None):
        self.classes_ = np.unique(y)
        self.n_labels = len(self.classes_)

        if self.n_random_features is not None:
            n_random_features = min(self.n_random_features, X.shape[1])
            features_subset = np.random.choice(X.shape[1], n_random_features, replace=False)
        else:
            features_subset = list(range(X.shape[1]))

        error, self.feature_idx, self.split_value, self.left_label, self.right_label = fit_stump_samme(
            X, y, features_subset, self.n_labels, sample_weight=sample_weights
        )

        return self

    def decision_function(self, X):
        prediction = np.zeros((X.shape[0], self.n_labels))
        prediction.fill(-1 / (self.n_labels - 1))
        mask = X[:, self.feature_idx] < self.split_value

        prediction[mask, self.left_label] = 1
        prediction[~mask, self.right_label] = 1
        return prediction

    def predict(self, X):
        return self.decision_function(X).argmax(axis=1)

    def __lt__(self, other):
        return self.split_value < other.split_value


class DecisionStumpSammeR(BaseEstimator, ClassifierMixin):
    def __init__(self, n_random_features=None):
        self.feature_idx = -1
        self.left_probabilities = None
        self.right_probabilities = None
        self.split_value = 0
        self.n_labels = 0
        self.n_random_features = n_random_features
        self.classes_ = 0

    def fit(self, X, y, sample_weight=None):
        self.classes_ = np.unique(y)
        self.n_labels = len(self.classes_)

        if self.n_random_features is not None:
            n_random_features = min(self.n_random_features, X.shape[1])
            features_subset = np.random.choice(X.shape[1], n_random_features, replace=False)
        else:
            features_subset = list(range(X.shape[1]))

        error, self.feature_idx, self.split_value, self.left_probabilities, self.right_probabilities = fit_stump_sammer(
            X, y, features_subset, self.n_labels, sample_weight=sample_weight
        )

        return self

    def predict_proba(self, X):
        probability = np.zeros((X.shape[0], self.n_labels))
        mask = X[:, self.feature_idx] < self.split_value

        probability[mask, :] = self.left_probabilities
        probability[~mask, :] = self.right_probabilities

        np.clip(probability, np.finfo(probability.dtype).eps, None, out=probability)

        return probability

    def predict(self, X):
        return self.predict_proba(X).argmax(axis=1)

    def decision_function(self, X):
        probabilities = self.predict_proba(X)

        np.clip(probabilities, np.finfo(probabilities.dtype).eps, None, out=probabilities)
        log_proba = np.log(probabilities)

        return (self.n_labels - 1) * (log_proba - (1. / self.n_labels) * log_proba.sum(axis=1)[:, np.newaxis])

    def __lt__(self, other):
        return self.split_value < other.split_value



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
from os.path import join
from os import walk

try:
    long_description = open("README.rst").read()
except IOError:
    long_description = ""

c_files = []
for (dirpath, dirs, files) in walk('valiml'):
    for fname in files:
        if (fname.endswith('.c') or fname.endswith('.h')) and not fname.startswith('_cffi_'):
            c_files.append(join(dirpath, fname))

setup(
    name="valiml",
    version="0.1.4",
    description="Extension of sklearn with tweaked implementation of common machine learning algorithms for self-use.",
    license="MIT",
    author="Valentin-Bogdan Roșca",
    packages=find_packages(),
    long_description=long_description,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
    ],
    setup_requires=["cffi"],
    install_requires=["cffi"],
    cffi_modules=[join('valiml', 'utils', 'src', 'cffi.py:ffibuilder')],
    package_data={
        '': ['*.h', '*.c']
    },
    include_package_data=True,
    data_files=[('.', c_files)],
    zip_safe=True
)



import importlib

import numba
import numpy as np
from numba import types
from numba.typed import Dict
from numba.cffi_support import register_module

loaded = importlib.import_module('_utils')
register_module(loaded)

get_seconds_since_epoch = loaded.lib.get_seconds_since_epoch
print_char_int = loaded.lib.print_char
print_float = loaded.lib.print_float

print_bar = loaded.lib.print_bar
print_info_bar = loaded.lib.print_info_bar
print_remaining = loaded.lib.print_remaining


@numba.jit(nopython=True)
def numba_ord(chr):
    if chr == '=':
        return 61
    elif chr == '.':
        return 46
    elif chr == '>':
        return 62
    elif chr == '[':
        return 91
    elif chr == ']':
        return 93
    elif chr == '\r':
        return 13
    elif chr == '\b':
        return 8
    elif chr == '\n':
        return 10
    elif chr == '_':
        return 95
    elif chr == ' ':
        return 32
    elif chr == '/':
        return 47
    elif chr == '+':
        return 43
    elif chr == '-':
        return 45
    elif chr == '*':
        return 42
    elif chr == ':':
        return 58
    elif chr == '{':
        return 123
    elif chr == '}':
        return 125
    elif chr == '0':
        return 48
    elif chr == '1':
        return 49
    elif chr == '2':
        return 50
    elif chr == '3':
        return 51
    elif chr == '4':
        return 52
    elif chr == '5':
        return 53
    elif chr == '6':
        return 54
    elif chr == '7':
        return 55
    elif chr == '8':
        return 56
    elif chr == '9':
        return 57
    elif chr == 'a':
        return 97
    elif chr == 'b':
        return 98
    elif chr == 'c':
        return 99
    elif chr == 'd':
        return 100
    elif chr == 'e':
        return 101
    elif chr == 'f':
        return 102
    elif chr == 'g':
        return 103
    elif chr == 'h':
        return 104
    elif chr == 'i':
        return 105
    elif chr == 'j':
        return 106
    elif chr == 'k':
        return 107
    elif chr == 'l':
        return 108
    elif chr == 'm':
        return 109
    elif chr == 'n':
        return 110
    elif chr == 'o':
        return 111
    elif chr == 'p':
        return 112
    elif chr == 'q':
        return 113
    elif chr == 'r':
        return 114
    elif chr == 's':
        return 115
    elif chr == 't':
        return 116
    elif chr == 'u':
        return 117
    elif chr == 'v':
        return 118
    elif chr == 'w':
        return 119
    elif chr == 'x':
        return 120
    elif chr == 'y':
        return 121
    elif chr == 'z':
        return 122
    elif chr == 'A':
        return 65
    elif chr == 'B':
        return 66
    elif chr == 'C':
        return 67
    elif chr == 'D':
        return 68
    elif chr == 'E':
        return 69
    elif chr == 'F':
        return 70
    elif chr == 'G':
        return 71
    elif chr == 'H':
        return 72
    elif chr == 'I':
        return 73
    elif chr == 'J':
        return 74
    elif chr == 'K':
        return 75
    elif chr == 'L':
        return 76
    elif chr == 'M':
        return 77
    elif chr == 'N':
        return 78
    elif chr == 'O':
        return 79
    elif chr == 'P':
        return 80
    elif chr == 'Q':
        return 81
    elif chr == 'R':
        return 82
    elif chr == 'S':
        return 83
    elif chr == 'T':
        return 84
    elif chr == 'U':
        return 85
    elif chr == 'V':
        return 86
    elif chr == 'W':
        return 87
    elif chr == 'X':
        return 88
    elif chr == 'Y':
        return 89
    elif chr == 'Z':
        return 90
    return -1


@numba.jit(nopython=True)
def print_string(s):
    for idx in range(len(s)):
        character = s[idx]
        print_char_int(numba_ord(character))

    return len(s)


@numba.jitclass([
    ('target', types.int32),
    ('width', types.int32),
    ('verbose', types.int8),
    ('interval', types.float32),
    ('_dynamic_display', types.boolean),
    ('_metrics_sum', Dict.empty(
        key_type=types.unicode_type,
        value_type=types.float64,
    )._numba_type_),
    ('_metrics_count', Dict.empty(
        key_type=types.unicode_type,
        value_type=types.int32,
    )._numba_type_),
    ('_total_width', types.int32),
    ('_seen_so_far', types.int32),
    ('_start', types.float64),
    ('_last_update', types.float64)
])
class ProgressBar:
    def __init__(self, target, width=30, verbose=1, interval=0.05,
                 stateful_metrics=None, dynamic_display=True):
        """
        Displays a pretty progress bar on screen. Inspired from keras.utils.Progbar

        :param target: The total number of steps expected, None if unknown.
        :param width: Progress bar width on screen.
        :param verbose: Verbosity: 0 (silent), 1 (verbose)
        :param interval: Minimum visual progress update interval (in seconds).
        :param stateful_metrics: Iterable of names of the metrics that should not be averaged over time.

        `>>> bar = ProgresBar(10, stateful_metrics=['error', 'cpu_usage'])`
        """
        self.target = -1 if target is None else target
        self.width = width
        self.verbose = verbose
        self.interval = interval

        self._metrics_sum = Dict.empty(
            key_type=types.unicode_type,
            value_type=types.float64
        )
        self._metrics_count = Dict.empty(
            key_type=types.unicode_type,
            value_type=types.int32
        )

        for metric_name in stateful_metrics.split('|'):
            if metric_name != '':
                self._metrics_sum[metric_name] = 0
                self._metrics_count[metric_name] = -1

        self._dynamic_display = dynamic_display
        self._total_width = 0
        self._seen_so_far = 0
        self._start = get_seconds_since_epoch()
        self._last_update = 0

    def update_metric(self, current, metric_name, metric_value):
        if metric_name not in self._metrics_count:
            self._metrics_sum[metric_name] = metric_value
            self._metrics_count[metric_name] = current - self._seen_so_far
        else:
            metric_count = self._metrics_count[metric_name]
            if metric_count == -1:
                self._metrics_sum[metric_name] = metric_value
            else:
                metric_sum = self._metrics_sum[metric_name]
                self._metrics_sum[metric_name] = metric_sum + metric_value
                self._metrics_count[metric_name] = metric_count + current - self._seen_so_far

    def update_metrics(self, current, metrics="", values=None):
        if values is None:
            return

        for metric, value in zip(metrics.split("|"), values):
            self.update_metric(current, metric, value)

    def print_metric(self, metric_name):
        characters_printed = 0

        characters_printed += print_string(' - ')
        characters_printed += print_string(metric_name)
        characters_printed += print_string(':')

        metric_sum = self._metrics_sum[metric_name]
        metric_count = self._metrics_count[metric_name]

        if metric_count > -1:
            avg = metric_sum / max(1., metric_count)
            if np.abs(avg) > 1e-3:
                characters_printed += print_float(avg, 4)
            else:
                characters_printed += print_float(avg, 4)
        else:
            characters_printed += print_string(' ')
            characters_printed += print_float(metric_sum, 4)
        return characters_printed

    def print_metrics(self):
        characters_printed = 0
        for metric_name in self._metrics_sum:
            characters_printed += self.print_metric(metric_name)
        return characters_printed

    def update(self, current: int, metrics="", values=None):
        """
        Update the current step and values.

        :param current: The current step index.
        :param values: The values to update.
        :return:
        """
        self.update_metrics(current, metrics, values)
        self._seen_so_far = current

        now = get_seconds_since_epoch()
        elapsed_from_start = now - self._start
        elapsed_from_last_update = now - self._last_update

        if self.verbose == 1:
            if elapsed_from_last_update < self.interval and self.target != -1 and current < self.target:
                return

            prev_total_width = self._total_width
            self._total_width = print_bar(current, self.target, prev_total_width, self.width, self._dynamic_display)

            info_len = print_info_bar(current, self.target, elapsed_from_start)
            info_len += self.print_metrics()

            self._total_width += info_len

            print_remaining(current, self.target, prev_total_width, self._total_width)

        self._last_update = now

    def add(self, n, metrics="", values=None):
        self.update(self._seen_so_far + n, metrics, values)


@numba.jit(nopython=True, cache=True)
def create_progbar_numba(target, width=30, verbose=1, interval=0.05, stateful_metrics="", dynamic_display=True):
    return ProgressBar(
        target,
        width,
        verbose,
        interval,
        stateful_metrics,
        dynamic_display
    )



from valiml.optimizers import lm
import numpy as np
from functools import partial

from valiml.utils.numba_utils import create_progbar_numba


class LSTSQ():
    def __init__(self, l2_alpha=None, constraints=None):
        self.l2_alpha = l2_alpha
        self.constraints = constraints
    
    def fit(self, X, Y):
        if self.l2_alpha is not None:
            l2_x = np.sqrt(self.l2_alpha) * np.identity(X.shape[1])
            l2_y = np.zeros((X.shape[1]))
            X = np.concatenate([X, l2_x])
            Y = np.concatenate([Y, l2_y])
                    
        if self.constraints is None:
            self.x = np.linalg.lstsq(X, Y, rcond=None)[0]
        else:
            zero_matrix = np.zeros((self.constraints[0].shape[0], self.constraints[0].shape[0]))
            Y = np.hstack([2 * np.matmul(X.T, Y), self.constraints[1]])
            X = np.block([
                [np.matmul(X.T, X), self.constraints[0].T],
                [self.constraints[0], zero_matrix]
            ])
            
            self.x = np.linalg.solve(X, Y)[:-self.constraints[1].shape[0]]
        
        return self
    
    def predict(self, X):
        return np.matmul(X, self.x)

    
class SigmaLSTSQ():
    def __init__(self, n_iters=1):
        self.n_iters = n_iters
    
    def predict(self, X):
        k = np.matmul(X, self.x)
        return k
    
    def _callback(self, X, Y, n_iter, x, y, alpha):
        self.x = x
        error = np.sum(np.sign(self.predict(X)) != Y) / Y.shape[0]

        metrics = "|".join(['alpha', 'loss', 'error', 'acc'])
        values = [alpha, y, error, 1 - error]
        self._bar.update(n_iter, metrics, values)
    
    def _sigma_derivative(self, x):
        s = np.exp(2 * x)
        return (4 * s) / (1 + s)**2

    def mse(self, X, Y, x, jacobian=True):
        z = np.matmul(X, x)
        y = np.tanh(z)
        loss = np.array([np.linalg.norm(y - Y)])**2

        if jacobian:
            derivative = np.matmul(2 * (y - Y) * np.vectorize(self._sigma_derivative)(z), X)
            return loss, derivative
        return loss

    def _fit(self, X, Y, l2_alpha=1e3, alpha=1e10, n_iters=20, eps=1e-6, verbose=True):        
        self._bar = create_progbar_numba(target=n_iters, stateful_metrics="|".join(['alpha', 'loss', 'error']))
        
        self.x = np.random.normal(scale=1/np.sqrt(X.shape[1]), size=X.shape[1])
        
        self.x = lm(self.x,
                    partial(self.mse, X, Y),
                    alpha=alpha,
                    max_iters=n_iters,
                    eps=eps,
                    callback=partial(self._callback, X, Y)
                    )
            
        return self
    
    def fit(self, X, Y, **kwargs):
        best_error = float('inf')
        best_x = None
        for n_iter in range(self.n_iters):
            self._fit(X, Y, **kwargs)
            
            error = (np.sign(self.predict(X)) != Y).sum() / Y.shape[0]
            if error < best_error:
                best_x = self.x
                best_error = error
        self.x = best_x
        print(f'Selecting best classifier with error {best_error}')
        return self



import sys


def normalize(x):
    return x / x.sum()


def create_progbar(target, width=30, verbose=1, interval=0.05, stateful_metrics=None, numba=True):
    if numba:
        from valiml.utils.numba_utils import ProgressBar

        dynamic_display = ((hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()) or 'ipykernel' in sys.modules)
        stateful_metrics = '|'.join(set(stateful_metrics)) if stateful_metrics else ''

        return ProgressBar(target,
                           width=width,
                           verbose=verbose,
                           interval=interval,
                           stateful_metrics=stateful_metrics,
                           dynamic_display=dynamic_display
                           )
